import os, logging
import numpy as np
from trview import ufiles

logger = logging.getLogger('trview.write_u')

lbl_d = { \
    'ip'    : ['SM', 'CUR'   , 'Plasma Current'.ljust(20) + 'Amps'], \
    'bt'    : ['S' , 'RBZ'   , 'Rp*Bt'.ljust(20)          + 'T.cm'], \
    'uloop' : ['U' , 'VSF'   , 'Vloop'.ljust(20)          + 'V'   ], \
    'prad'  : ['P' , 'RAD'   , 'Prad'.ljust(20)           + 'W'   ], \
    'betpol': ['B' , 'BETPOL', 'Beta poloidal'.ljust(20)          ], \
    'zef'   : ['Z' , 'ZEF'   , 'Effective charge'.ljust(20)       ] }

tlbl = 'Time'.ljust(20) + 'Seconds'

def as_ubnd(fit_sep, udir=None):

    if hasattr(fit_sep, 'time'):
        ind_t = fit_sep.ind_ok
        t_u  = fit_sep.time[fit_sep.ind_ok]
    else: #eqdsk
        ind_t = [0]
        t_u = np.array([1.])
    Rbnd = np.atleast_2d(fit_sep.r_bnd)[ind_t, :-1]
    Zbnd = np.atleast_2d(fit_sep.z_bnd)[ind_t, :-1]

    nt, n_the = Rbnd.shape
    posarr = 1 + np.arange(n_the)
    poslbl = 'POSITION'
    bndlbl = 'R BOUNDARY    [M]'

    uf_d = { 'pre': '', 'ext': 'bnd.txt_r', 'shot': fit_sep.shot, \
             'grid': { 'X':{'lbl': tlbl  , 'arr': t_u}, \
                       'Y':{'lbl': poslbl, 'arr': posarr} }, \
             'data': {'lbl': bndlbl, 'arr': Rbnd} }
    ufiles.WU(uf_d, udir=udir)

    uf_d = { 'pre': '', 'ext': 'bnd.txt_z', 'shot': fit_sep.shot, \
             'grid': { 'X':{'lbl': tlbl  , 'arr': t_u}, \
                       'Y':{'lbl': poslbl, 'arr': posarr} }, \
             'data': {'lbl': bndlbl, 'arr': Zbnd} }
    ufiles.WU(uf_d, udir=udir)
    

def tr_u_geo(fit_sep):

    t_u = fit_sep.time[fit_sep.ind_ok]
    moments_u = 100.*fit_sep.moments[fit_sep.ind_ok, :, :]

    nt, mom_order, mom_type = moments_u.shape
    marr = 1 + np.arange(mom_order)
    karr = 1 + np.arange(mom_type)

    mlbl = 'MOMENT INDEX'
    klbl = 'MOMENT TYPE'
    dlbl = 'FOURIER MOMENTS'.ljust(20) + 'CM'

    uf_d = { 'pre': 'M', 'ext': 'MRY', 'shot': fit_sep.shot, \
             'grid': { 'X':{'lbl': tlbl, 'arr': t_u}, \
                       'Y':{'lbl': mlbl, 'arr': marr}, \
                       'Z':{'lbl': klbl, 'arr': karr} }, \
             'data':       {'lbl': dlbl, 'arr': moments_u} }
    uf_d['comm'] = 'exp=%s\n diag=%s\n ed=%d' %(fit_sep.exp, fit_sep.diag, fit_sep.ed)

    ufiles.WU(uf_d)


def write_q(equ, udir=None):

    rholbl = 'rho_tor'
    rho = equ.rho_tor_n[0, :]
    q   = np.abs(equ.q[0, :])
    time = equ.time[0]

    dlbl = 'Safety factor'
    uf_d = { 'pre': 'Q', 'ext': 'QPR', 'shot': equ.shot, \
             'grid': {'X': {'lbl': rholbl, 'arr': rho}}, \
             'data':       {'lbl': dlbl, 'arr': q} }
    uf_d['comm']  = 'time=%8.4f s\n' %time
    uf_d['comm'] += ' exp=%s\n diag=%s\n ed=%d' %(equ.exp, equ.diag, equ.ed)
    ufiles.WU(uf_d, udir=udir)


def write_q2d(equ, udir=None):

    rholbl = 'rho_tor'
    rho = equ.rho_tor_n[0, :-1]
    q   = np.atleast_2d(np.abs(equ.q[0, :-1]))
    time = np.atleast_1d(equ.time[0])

    dlbl = 'Safety factor'
    uf_d = { 'pre': 'Q', 'ext': 'QPR', 'shot': equ.shot, \
             'grid': {'X': {'lbl': tlbl  , 'arr': time}, \
                      'Y': {'lbl': rholbl, 'arr': rho}}, \
             'data':       {'lbl': dlbl, 'arr': q} }
    uf_d['comm'] = ' exp=%s\n diag=%s\n ed=%d' %(equ.exp, equ.diag, equ.ed)
    ufiles.WU(uf_d, udir=udir)


def write_u1d(sig1d, nbi, ech, icr, R0=165., code='tr', udir=None):

    if not hasattr(sig1d, 'ip'):
        logger.error('Missing current, no 1D u-files written')
        return

    for key, data in sig1d.__dict__.items():
        if key in lbl_d.keys():
            str_u = lbl_d[key]
            pre  = str_u[0]
            ext  = str_u[1]
            dlbl = str_u[2]
            t_u = data.time
            d_u = data.data.copy()

            sign = np.sign(np.average(d_u))

            if code == 'as':
                if key == 'bt':
                    d_u *= -1
                    pre = 'MAG'
                    ext = 'BTOR'
                    dlbl = 'Btor'.ljust(29) + 'T'
                elif key == 'ip':
                    d_u *= 1e-6
                    pre = 'MAG'
                    ext = 'IPL'
                    dlbl = 'Plasma Current'.ljust(20) + 'MA'
                else:
                    continue
            else:
                if key == 'bt':
                    d_u *= -R0 # major radius [cm]

            if code == 'tr' and key in ('ip', 'bt'):
                d_u = np.abs(d_u)

            d_u[0] = np.abs(d_u[0])

            uf_d = { 'pre': pre, 'ext': ext, 'shot': data.shot, \
                     'grid': {'X': {'lbl': tlbl, 'arr': t_u}}, \
                     'data':       {'lbl': dlbl, 'arr': d_u} }
            comm = ' AUGD:%s:%d' %(data.diagsig, data.edition)
            uf_d['comm'] = comm
            ufiles.WU(uf_d, udir=udir)

            t_av = np.atleast_1d(np.nanmean(t_u))
            d_av = np.atleast_1d(np.nanmean(d_u))
            ufa_d = { 'pre': pre, 'ext': ext + '_AVG', 'shot': data.shot, \
                      'grid': {'X': {'lbl': tlbl, 'arr': t_av}}, \
                      'data':       {'lbl': dlbl, 'arr': d_av} }
            ufa_d['comm'] = ' Time averaged %8.4f - %8.4f\n' %(t_u[0], t_u[-1])
            ufa_d['comm'] += comm
            ufiles.WU(ufa_d, udir=udir)

# NBI

    if hasattr(nbi, 'pnbi_red'):

        nsrc = nbi.pnbi_red.shape[1]

        xlbl = 'Channel number'
        dlbl = 'NBI power'.ljust(20) + 'W'
        x_nbi = 1. + np.arange(nsrc)
        t_u = nbi.t_nbi_red
        d_u = nbi.pnbi_red
        if np.max(nbi.pnbi_red) > nbi.nbi_min:
            uf_2d = { 'pre': 'P', 'ext': 'NBI', 'shot': nbi.shot, \
                     'grid': {'X': {'lbl': tlbl, 'arr': t_u}, \
                              'Y': {'lbl': xlbl, 'arr': x_nbi}}, \
                     'data':       {'lbl': dlbl, 'arr': d_u} }
            ufiles.WU(uf_2d, udir=udir)

            t_av = np.atleast_1d(np.nanmean(t_u))
            d_av = np.atleast_2d(np.nanmean(d_u, axis=0))
            ufa_d = { 'pre': 'P', 'ext': 'NBI_AVG', 'shot': nbi.shot, \
                      'grid': {'X': {'lbl': tlbl, 'arr': t_av}, \
                               'Y': {'lbl': xlbl, 'arr': x_nbi}}, \
                      'data':       {'lbl': dlbl, 'arr': d_av} }
            ufa_d['comm'] = ' Time averaged %8.4f - %8.4f' %(t_u[0], t_u[-1])
            ufiles.WU(ufa_d, udir=udir)

# ECRH

    if hasattr(ech, 'pech_ds'):

        xlbl = 'Channel number'
        dlbl = 'ECRH power'.ljust(20) + 'W'
        ngy = len(ech.freq)
        x_ech = 1. + np.arange(ngy)
        t_u = ech.t_ds
        d_u = ech.pech_ds
        if code == 'tr':
            prephi = 'PHITR'
            prethe = 'THETR'
            phi_fac = -1
            the_fac = -1
        elif code == 'as':
            prephi = 'PHIAS'
            prethe = 'THEAS'
            phi_fac = 1
            the_fac = 1
            
        if np.max(ech.pech_ds) > ech.ech_min:
            uf_d = { 'pre': 'P', 'ext': 'ECH', 'shot': ech.shot, \
                     'grid': {'X': {'lbl': tlbl, 'arr': t_u}, \
                              'Y': {'lbl': xlbl, 'arr': x_ech} }, \
                     'data':       {'lbl': dlbl, 'arr': d_u}   }
            ufiles.WU(uf_d, udir=udir)

            t_av = np.atleast_1d(np.nanmean(t_u))
            d_av = np.atleast_2d(np.nanmean(d_u, axis=0))
            ufa_d = { 'pre': 'P', 'ext': 'ECH_AVG', 'shot': ech.shot, \
                      'grid': {'X': {'lbl': tlbl, 'arr': t_av}, \
                               'Y': {'lbl': xlbl, 'arr': x_ech}}, \
                      'data':       {'lbl': dlbl, 'arr': d_av} }
            ufa_d['comm'] = ' Time averaged %8.4f - %8.4f' %(t_u[0], t_u[-1])
            ufiles.WU(ufa_d, udir=udir)

# Angles

    if hasattr(ech, 'phi_ecs'):

        t_ang = ech.t_c_ecs
        phi   = ech.phi_ecs
        the   = ech.the_ecs

    elif hasattr(ech, 'phi_ds'):

        t_ang = ech.t_ang_ds
        phi   = ech.phi_ds
        the   = ech.the_ds

    if 't_ang' in locals():

        t_u = t_ang
        dlbl = 'ECRH tor. angle'.ljust(20) + 'deg'
        uf_d = { 'pre': prephi, 'ext': 'ECH', 'shot': ech.shot, \
                'grid': {'X': {'lbl': tlbl, 'arr': t_u}, \
                         'Y': {'lbl': xlbl, 'arr': x_ech} }, \
                'data':       {'lbl': dlbl, 'arr': phi_fac*phi}   }
        ufiles.WU(uf_d, udir=udir)

        dlbl = 'ECRH pol. angle'.ljust(20) + 'deg'
        uf_d = { 'pre': prethe, 'ext': 'ECH', 'shot': ech.shot, \
                'grid': {'X': {'lbl': tlbl, 'arr': t_u}, \
                         'Y': {'lbl': xlbl, 'arr': x_ech} }, \
                'data':       {'lbl': dlbl, 'arr': the_fac*the}   }
        ufiles.WU(uf_d, udir=udir)

# Time averaged
        t_av   = np.atleast_1d(np.nanmean(t_u))
        phi_av = np.atleast_2d(np.nanmean(phi, axis=0))
        the_av = np.atleast_2d(np.nanmean(the, axis=0))

        dlbl = 'ECRH tor. angle'.ljust(20) + 'deg'
        ufa_d = { 'pre': prephi, 'ext': 'ECH_AVG', 'shot': ech.shot, \
                'grid': {'X': {'lbl': tlbl, 'arr': t_av}, \
                         'Y': {'lbl': xlbl, 'arr': x_ech} }, \
                'data':       {'lbl': dlbl, 'arr': phi_fac*phi_av}   }
        ufa_d['comm'] = ' Time averaged %8.4f - %8.4f' %(t_u[0], t_u[-1])
        ufiles.WU(ufa_d, udir=udir)

        dlbl = 'ECRH pol. angle'.ljust(20) + 'deg'
        ufa_d = { 'pre': prethe, 'ext': 'ECH_AVG', 'shot': ech.shot, \
                'grid': {'X': {'lbl': tlbl, 'arr': t_av}, \
                         'Y': {'lbl': xlbl, 'arr': x_ech} }, \
                'data':       {'lbl': dlbl, 'arr': the_fac*the_av}   }
        ufa_d['comm'] = ' Time averaged %8.4f - %8.4f' %(t_u[0], t_u[-1])
        ufiles.WU(ufa_d, udir=udir)

# ICRF

    if hasattr(icr, 'picr'):

        nant = icr.picr_red.shape[1]
        xlbl = 'Channel number'
        dlbl = 'ICRH power'.ljust(20) + 'W'
        x_icr = 1. + np.arange(nant)
        t_u = icr.t_icr_red
        d_u = icr.picr_red
        if np.max(icr.picr_red) > icr.icr_min:
            uf_d = { 'pre': 'P', 'ext': 'ICH', 'shot': icr.shot, \
                     'grid': {'X': {'lbl': tlbl, 'arr': t_u}, \
                              'Y': {'lbl': xlbl, 'arr': x_icr}}, \
                     'data':       {'lbl': dlbl, 'arr': d_u} }
            ufiles.WU(uf_d, udir=udir)

            t_av = np.atleast_1d(np.nanmean(t_u))
            d_av = np.atleast_2d(np.nanmean(d_u, axis=0))
            ufa_d = { 'pre': 'P', 'ext': 'ICH_AVG', 'shot': icr.shot, \
                      'grid': {'X': {'lbl': tlbl, 'arr': t_av}, \
                               'Y': {'lbl': xlbl, 'arr': x_icr}}, \
                      'data':       {'lbl': dlbl, 'arr': d_av} }
            ufa_d['comm'] = ' Time averaged %8.4f - %8.4f' %(t_u[0], t_u[-1])
            ufiles.WU(ufa_d, udir=udir)

    if hasattr(icr, 't_cxf'):

        dlbl = 'Minority concentration'

        if icr.minr is not None:
            uf_d = { 'pre': 'ICR', 'ext': 'MINR', 'shot': icr.shot, \
                     'grid': {'X': {'lbl': tlbl, 'arr': icr.t_cxf} }, \
                     'data':       {'lbl': dlbl, 'arr': icr.minr} }
            ufiles.WU(uf_d, udir=udir)

        if icr.minl is not None:
            uf_d = { 'pre': 'ICR', 'ext': 'MINL', 'shot': icr.shot, \
                     'grid': {'X': {'lbl': tlbl, 'arr': icr.t_cxf} }, \
                     'data':       {'lbl': dlbl, 'arr': icr.minl} }
            ufiles.WU(uf_d, udir=udir)


def write_u2d(pr_obj, coord='rho_tor', udir=None):

    xfac = 1.

    for prefix, prof in pr_obj.prof.items():

        t_u = pr_obj.time
        if coord == 'rho_tor':
            d_u = prof.yfit_rhot
        elif coord == 'rho_pol':
            d_u = prof.yfit_rhop
        if hasattr(prof, 'tnot'):
            t_u = np.delete(t_u, prof.tnot)
            d_u = np.delete(d_u, prof.tnot, axis=0)

        ext_str = ''
        for sflist in prof.sflists:
            diag = sflist.split(':')[1]
            ext_str += diag

# Find unit label and conversion factor
        f_labels = '%s/labels.dat' %os.path.dirname(os.path.realpath(__file__))
        prestr, dlab, tr_lbl, tr_unit, as_name, trfac, asfac = \
            np.genfromtxt(f_labels, dtype='str', skip_header=2, unpack=True)
        (j_pre, ) = np.where(prestr == prefix)
        if len(j_pre) > 0:
            jpre = j_pre[0]
            if prefix in ('E', 'I', 'N'):
                d_u = np.maximum(d_u, 1e-6)
            if tr_unit[jpre] != 'none':
                dlbl = '%-20s%-10s' %(dlab[jpre], tr_unit[jpre])
            else:
                dlbl = '%-30s' %dlab[jpre]
            tr_fac = float(trfac[jpre])
            logger.debug('%s %s %s %s', t_u.shape, pr_obj.rho.shape, d_u.shape, prof.units)

            if prefix == 'V' and prof.units == 'rad/s':
                prefix = 'ANGF'
                tr_fac = 1
                dlbl = 'Angular frequency'.ljust(20) + prof.units
            uf_d = { 'pre': prefix, 'ext': ext_str, 'shot': pr_obj.shot, \
                     'grid': {'X': {'lbl': tlbl     , 'arr': t_u}, \
                              'Y': {'lbl': pr_obj.xlbl, 'arr': xfac*pr_obj.rho} }, \
                     'data': {'lbl': dlbl, 'arr': tr_fac*d_u} }

            comm = 'Shotfiles:\n'
            for sfl in prof.sflists:
                comm += '  %s\n' %sfl
            comm += ' Mapping with equilibrium:\n'
            comm +=  '  exp=%s\n diag=%s\n ed=%s\n' %(pr_obj.eq_exp, pr_obj.eq_diag, pr_obj.eq_ed)
            if hasattr(prof, 'method'):
                comm += ' Method: %s\n' %prof.method
            if all(hasattr(prof, attr) for attr in ['fit_tol', 'sigmas']):
                comm += '  fit tolerance=%8.4f\n sigmas=%8.4f\n' %(prof.fit_tol, prof.sigmas)
            uf_d['comm'] = comm
            ufiles.WU(uf_d, udir=udir)

# Tiem averaged profile

            ext_avg = '%s_AVG' %(ext_str)
            t_av = np.atleast_1d(np.average(t_u))
            dat_av = tr_fac*np.average(d_u, axis=0).reshape((1, d_u.shape[1]))
            ufa_d = { 'pre': prefix, 'ext': ext_avg, 'shot': pr_obj.shot, \
                      'grid': {'X': {'lbl': tlbl     , 'arr': t_av}, \
                               'Y': {'lbl': pr_obj.xlbl, 'arr': xfac*pr_obj.rho}}, \
                      'data': {'lbl': dlbl, 'arr': dat_av} }

            ufa_d['comm'] = ' Time averaged %8.4f - %8.4f\n' %(t_u[0], t_u[-1])
            ufa_d['comm'] += comm

            ufiles.WU(ufa_d, udir=udir)
